"""
Constants having to do with Gatekeeper

.. data:: GATEKEEPER_KEY_ENVIRONMENT_NAME

    The name of the environment variable for the gatekeeper key

"""

GATEKEEPER_KEY_ENVIRONMENT_NAME: str = "EHELPLY_GATEKEEPER_KEY"
